import traceback
import uuid
from typing import Dict

from .models import (
    AggregatedException,
    ExceptionInstance,
    ExceptionWithContext,
    HTTPContext,
    Payload,
)


class ExceptionCollector:
    def __init__(self):
        self._aggregated_exceptions: Dict[str, AggregatedException] = {}
        self._uuid = str(uuid.uuid1())

    def report(self, exception: Exception):
        """
        Report an exception

        :param Exception exception: captured exception
        """
        self._add_exception(exception)

    def report_with_context(self, exception: Exception, http_context: HTTPContext):
        """
        Report an exception with the context of an HTTP request

        :param Exception exception: captured exception
        :param HTTPContext http_context: context of an HTTP request
        """
        self._add_exception(exception, http_context)

    def get_aggregated_exceptions(self) -> Payload:
        """
        Get all aggregated exceptions in a list format

        :return Payload: list of aggregated exceptions
        """
        return Payload(
            aggregated_errors=list(self._aggregated_exceptions.values()),
            target_uuid=self._uuid,
        )

    def _add_exception(self, exception: Exception, http_context: HTTPContext = None):
        stacktrace = traceback.format_exc().strip().split("\n")
        error_class = type(exception).__name__

        exception_instance = ExceptionInstance(
            cls=error_class, message=str(exception), stacktrace=stacktrace
        )
        exception_with_context = ExceptionWithContext(
            error=exception_instance, http_context=http_context
        )

        aggregation_key = exception_with_context.aggregation_key()
        if aggregation_key not in self._aggregated_exceptions:
            self._aggregated_exceptions[aggregation_key] = AggregatedException(
                aggregation_key=aggregation_key, latest_errors=[]
            )
        self._aggregated_exceptions[aggregation_key].add_exception(
            exception_with_context
        )



import requests
from .collector import ExceptionCollector


class ExceptionExporter:
    def __init__(self, collector: ExceptionCollector):
        self._collector = collector

    def export(self) -> str:
        """
        Export the collection of errors in JSON format

        :return: str
        """
        return self._collector.get_aggregated_exceptions().to_json()  # type: ignore

    def push_to_gateway(self, addr: str):
        requests.post(addr + "/errors", self.export())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['periskop_client']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses-json>=0.4.2,<0.5.0']

setup_kwargs = {
    'name': 'periskop-client',
    'version': '0.2.0',
    'description': 'Periskop Python client',
    'long_description': '# periskop-python\n[![pypi](https://img.shields.io/pypi/v/periskop-client.svg)](https://pypi.python.org/pypi/periskop-client/) [![PyPI pyversions](https://img.shields.io/pypi/pyversions/periskop-client.svg)](https://pypi.python.org/pypi/periskop-client/) [![Build Status](https://api.cirrus-ci.com/github/soundcloud/periskop-python.svg)](https://cirrus-ci.com/github/soundcloud/periskop-python)\n\n\n[Periskop](https://github.com/soundcloud/periskop) requires collecting and aggregating exceptions on the client side,\nas well as exposing them via an HTTP endpoint using a well defined format.\n\nThis library provides low level collection and rendering capabilities\n\n## Install\n\n```\npip install periskop-client\n```\n\n## Usage example\n\n```python\nimport json\nfrom http.server import HTTPServer\n\nfrom periskop_client.collector import ExceptionCollector\nfrom periskop_client.exporter import ExceptionExporter\nfrom periskop_client.handler import exception_http_handler\nfrom periskop_client.models import HTTPContext\n\n\ndef faulty_json():\n    return json.loads(\'{"id":\')\n\n\nif __name__ == "__main__":\n    collector = ExceptionCollector()\n    try:\n        faulty_json()\n    except Exception as exception:\n        # Report without context\n        collector.report(exception)\n        # Report with HTTP context without request body\n        collector.report_with_context(\n            exception,\n            HTTPContext("GET", "http://example.com", {"Cache-Control": "no-cache"}),\n        )\n        # Report with HTTP context with request body\n        collector.report_with_context(\n            exception,\n            HTTPContext("GET", "http://example.com", {"Cache-Control": "no-cache"}, "request_body"),\n        )\n\n    # Expose collected exceptions in localhost:8081/-/exceptions\n    server_address = ("", 8081)\n    handler = exception_http_handler(\n        path="/-/exceptions", exporter=ExceptionExporter(collector)\n    )\n    http_server = HTTPServer(server_address, handler)\n    http_server.serve_forever()\n```\n\n### Using push gateway\n\nYou can also use [pushgateway](https://github.com/soundcloud/periskop-pushgateway) in case you want to push your metrics \ninstead of using pull method. Use only in case you really need it (e.g a batch job) as it would degrade the performance \nof your application. In the following example, we assume that we deployed an instance of periskop-pushgateway \non `http://localhost:6767`:\n\n```scala\nexporter = ExceptionExporter(collector)\nexporter.push_to_gateway("http://localhost:6767")\n```\n\n## Run tests\n\nFor running tests [pytest](https://docs.pytest.org) is needed. A recommended way to run all check is installing [tox](https://tox.readthedocs.io/en/latest/install.html) and then just type `tox`. This will run `pytest` tests, [black](https://black.readthedocs.io) formatter and [flake8](https://flake8.pycqa.org) and [mypy](http://mypy-lang.org/) static code analyzers.\n\nAlternatively you can run `pip install -r requirements-tests.txt` and then run `pytest`.\n',
    'author': 'Marc Tuduri',
    'author_email': 'marc.tuduri@soundcloud.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import hashlib
import uuid
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List, Optional

from dataclasses_json import config, dataclass_json

SEVERITY_INFO = "info"
SEVERITY_WARNING = "warning"
SEVERITY_ERROR = "error"
MAX_TRACES = 5
MAX_ERRORS = 10
NUM_HASH_CHARS = 8


@dataclass_json
@dataclass
class ExceptionInstance:
    cls: str = field(metadata=config(field_name="class"))
    message: str
    stacktrace: List[str]
    # should be ExceptionInstance, but we don't needed in this lib
    cause: object = None


@dataclass_json
@dataclass
class HTTPContext:
    request_method: str
    request_url: str
    request_headers: Optional[Dict[str, str]] = None
    request_body: Optional[str] = None


@dataclass_json
@dataclass
class ExceptionWithContext:
    error: ExceptionInstance
    http_context: Optional[HTTPContext] = None
    severity: str = SEVERITY_ERROR
    uuid: str = ""
    timestamp: str = ""

    def __post_init__(self):
        self.uuid = str(uuid.uuid1())
        self.timestamp = datetime.utcnow().isoformat(timespec="milliseconds") + "Z"

    def _hash_exception(self, exception: str):
        h = hashlib.md5(exception.encode())
        return h.hexdigest()[:NUM_HASH_CHARS]

    def aggregation_key(self) -> str:
        """
        Generates the aggregation key with a hash using the last MAX_TRACES

        :return: str
        """
        stacktrace_head = self.error.stacktrace[:]
        if len(self.error.stacktrace) > MAX_TRACES:
            stacktrace_head = stacktrace_head[:MAX_TRACES]
        error_hash = self._hash_exception("".join(stacktrace_head))
        return f"{self.error.cls}@{error_hash}"


@dataclass_json
@dataclass
class AggregatedException:
    aggregation_key: str
    latest_errors: List[ExceptionWithContext]
    total_count: int = 0
    severity: str = SEVERITY_ERROR
    created_at: str = ""

    def __post_init__(self):
        self.created_at = datetime.utcnow().isoformat(timespec="milliseconds") + "Z"

    def add_exception(self, exception_with_context: ExceptionWithContext):
        """
        Add exception to the list of latest errors up to MAX_ERRORS

        :param ExceptionWithContext exception_with_context: exception with context that is being reported
        """
        if len(self.latest_errors) >= MAX_ERRORS:
            self.latest_errors.pop()
        self.latest_errors.append(exception_with_context)
        self.total_count += 1


@dataclass_json
@dataclass
class Payload:
    aggregated_errors: List[AggregatedException]
    target_uuid: str



from http.server import BaseHTTPRequestHandler

from .exporter import ExceptionExporter


def exception_http_handler(path: str, exporter: ExceptionExporter):
    """
    Factory function to create an HTTP Handler wraped with a Periskop exporter.

    :param str path: path to server exported exceptions
    :param ExceptionExporter exporter: exporter object to use as expose a JSON

    :return ExceptionBaseHTTPRequestHandler:
    """

    class ExceptionBaseHTTPRequestHandler(BaseHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            self._exporter = exporter
            super().__init__(*args, **kwargs)

        def _set_headers(self):
            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()

        def do_GET(self):
            if self.path == path:
                self._set_headers()
                self.wfile.write(self._exporter.export().encode(encoding="utf_8"))

    return ExceptionBaseHTTPRequestHandler
